import traceback

from django.conf import settings
from django.contrib.auth.models import User
from django.db.models import Manager, Model, QuerySet
from django.db.models.base import ModelBase

from djangotools.common import response, BaseSerializationModel, redirect
from djangotools.common.errors import CustomException
from djangotools.common.path import PointPath, ModelPath, AttributePath, ConstPath, VariablePath
from djangotools.common.loggin import debug_login



class Route:
    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    PUT = "PUT"
    method  = {GET, POST, DELETE, PUT}
    def __init__(self, route=None,  method=None, **options):
        self.options = options
        self.method = method if method is not None else self.method
        if isinstance(method, str): self.method = {method}
        if isinstance(route, str):
            route = ConstPath(route)
        self.path = route or PointPath()

    def __repr__(self):
        return f"<Route {self.method} {self.path.as_string()} >"

    def __call__(self, *args, **kwargs):
        return RegisteredRoute(self, args[0])

    def __truediv__(self, other):
        if isinstance(other, Route):
            return Route(self.path / other.path, method=other.method)
        if isinstance(other, RegisteredRoute):
            return RegisteredRoute(self / other.route, other.callback)
        raise Exception()


def request_wrapper(fct):
    def wrapper(request, *args, **kwargs):
        try:
            return fct(request, *args, **kwargs)
        except CustomException as err:
            return err.get_error()
        except Exception as err:
            traceback.print_exc()
            return response.serv_json(500, 500, err.__class__.__name__, str(err))

    return  wrapper

class RegisteredRoute:


    _default_user_id_ = 1
    def __init__(self, path : Route, callback, is_manager=False, is_method=True):
        """

        :param path:
        :param callback:
        :param is_manager:
        :param is_method:
        :param options:
            need_auth: active l'authentitficatyion
            no_wrap: ne wrappe pas les callbacks (pour les captures d'exception)
        """
        self.route = path if path is not None else Route()

        self.is_manager = is_manager
        self.is_method = is_method
        self.options = self.route.options
        if not "no_wrap" in self.options: callback = request_wrapper(callback)
        self.callback = callback


    def __repr__(self):
        return f"<RegisteredRoute {self.route} >"

    def _call(self, req, *args, **kwargs):
        obj = None
        debug_login(req, self._default_user_id_)
        user = req.user
        if self.options.get("need_auth", False) and  not user.is_authenticated:
            if self.options.get("redirect_login", False):
                return response.serv_json_not_logged()
            else:
                url = settings.LOGIN_URL if settings.LOGIN_URL.startswith("/") else "/"+settings.LOGIN_URL
                return redirect(url)

        for part in self.route.path.iter_models():
            if obj is None:
                obj = part.model.objects

            if isinstance(part, ModelPath):
                if not isinstance(obj, Manager):
                    obj = part.model.objects
            elif isinstance(part, AttributePath):
                obj = obj.filter(**{part.attr: kwargs[part.key]})
                if len(obj) > 2:
                    return response.serv_json_bad_request(f"Trop de résultats pour {part.model_name}{part.attr} = {kwargs[part.key]}")
                elif len(obj) == 0:
                    return response.serv_json_not_found(f"Aucun résultat pour {part.model_name}.{part.attr} = {kwargs[part.key]}")
                else:
                    kwargs.pop(part.key)
                    obj = obj.first()
                    if part.forward:
                        obj = getattr(obj, part.forward)
            elif isinstance(part, VariablePath):
                kwargs[part.key] = part.resolve_var(obj, kwargs)
            else:
                raise Exception()


        if self.is_manager:
            if isinstance(obj, Model):
                obj = obj.__class__.objects
            elif isinstance(obj, ModelBase):
                obj = obj.objects
        if self.is_method:
            ret = self.callback(obj, req, *args, **kwargs)
        else:
            if obj is None:
                ret = self.callback(req, *args, **kwargs)
            else:
                ret = self.callback(req, obj, *args, **kwargs)

        if isinstance(ret, QuerySet):
            return [x.serialize() for x in ret]
        if isinstance(ret, BaseSerializationModel):
            return ret.serialize()
        if isinstance(ret, list):
            ret = [x.serialize() if isinstance(x, BaseSerializationModel) else x for x in ret]
        return ret


    def __call__(self, req, *args, **kwargs):
        if "no_wrap" in self.options:
            return self._call(req, *args, **kwargs)
        return request_wrapper(self._call)(req, *args, **kwargs)


    def copy(self):
        return RegisteredRoute(self.route, self.callback, is_manager=self.is_manager,
                               is_method=self.is_method)

class Get(Route):
    method = {Route.GET}
    
class Post(Route):
    method = {Route.POST}

class Put(Route):
    method = {Route.PUT}

class Delete(Route):
    method = {Route.DELETE}
